<?php

namespace Instant\Credit\Api;

use Magento\Setup\Exception;

/**
 * Class AbstractApi
 * @package Instant\Credit\Api
 * @author Orenyi
 */
abstract class AbstractApi
{

    /**
     * @var \Magento\Framework\App\Request\Http
     */
    protected $request;

    /**
     * @var \Magento\Framework\App\Config\ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var \Magento\Framework\Serialize\Serializer\Json
     */
    protected $serializer;

    /**
     * @var \Magento\Framework\Webapi\Rest\Response
     */
    protected $apiResponse;

    /**
     * @var \Magento\Framework\HTTP\Client\Curl
     */
    protected $curl;

    /**
     * @var \Instant\Credit\Logger\InstantCredit
     */
    public $logger;

    /**
     * @var \Instant\Credit\Api\Data\TokenInterfaceFactory
     */
    protected $token;

    const API_ERROR_CODE_SUCCESS      =   '200';
    const API_ERROR_ACCESS_DENIED     =   '401';
    const API_ERROR_CODE_ERROR        =   '500';
    const DEFAULT_RESPONSE_MESSAGE    =   '';

    const TEST_API_URL = 'https://test.instantcredit.net/api/';
    const PRODUCTION_API_URL = 'https://api.instantcredit.net/api/';

    /**
     * @param \Magento\Framework\App\Request\Http $request
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Magento\Framework\Serialize\Serializer\Json $serializer
     * @param \Magento\Framework\Webapi\Rest\Response $apiResponse
     * @param \Magento\Framework\HTTP\Client\Curl $curl
     * @param \Instant\Credit\Logger\InstantCredit $logger
     * @param \Instant\Credit\Api\Data\TokenInterfaceFactory $token
     */
    public function __construct(
        \Magento\Framework\App\Request\Http $request,
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Magento\Framework\Serialize\Serializer\Json $serializer,
        \Magento\Framework\Webapi\Rest\Response $apiResponse,
        \Magento\Framework\HTTP\Client\Curl $curl,
        \Instant\Credit\Logger\InstantCredit $logger,
        \Instant\Credit\Api\Data\TokenInterfaceFactory $token
    ) {
        $this->request = $request;
        $this->scopeConfig = $scopeConfig;
        $this->serializer = $serializer;
        $this->apiResponse = $apiResponse;
        $this->curl = $curl;
        $this->token = $token;
        $this->logger = $logger;
    }

    /**
     * @param $data
     * @param $errorCode
     * @param string $msg
     * @return array[]
     */
    protected function createResponse($data, $errorCode, $msg = self::DEFAULT_RESPONSE_MESSAGE)
    {
        return [
            'response' => [
                'code' => $errorCode,
                'message' => (empty($msg) && empty($data)) ? 'No data found' : $msg,
                'data' => $data
            ]
        ];
    }

    /**
     * @param $data
     * @param string $msg
     * @return array[]
     */
    protected function createErrorResponse($data, $msg = self::DEFAULT_RESPONSE_MESSAGE)
    {
        $this->apiResponse->setHttpResponseCode(self::API_ERROR_CODE_ERROR);
        return $this->createResponse($data, self::API_ERROR_CODE_ERROR, $msg);
    }

    /**
     * @param $data
     * @param string $msg
     * @return array[]
     */
    public function createSuccessResponse($data, $msg = self::DEFAULT_RESPONSE_MESSAGE)
    {
        return $this->createResponse($data, self::API_ERROR_CODE_SUCCESS, $msg);
    }

    /**
     * @param null $msg
     * @return array[]
     */
    protected function createDeniedResponse($msg = null)
    {
        $this->apiResponse->setHttpResponseCode(self::API_ERROR_ACCESS_DENIED);
        return $this->createResponse(
            [],
            self::API_ERROR_ACCESS_DENIED,
            !is_null($msg) ? $msg : __('Invalid credentials to access this resource')
        );
    }

    /**
     * @param $configPath
     * @return mixed
     */
    public function getConfigValue($configPath)
    {
        return $this->scopeConfig->getValue($configPath, \Magento\Store\Model\ScopeInterface::SCOPE_STORE);
    }

    /**
     * @param $expectedParams
     * @param $data
     * @throws \Instant\Credit\Api\ApiException
     */
    public function checkBodyParameters($expectedParams, $data)
    {
        $paramsNotFound = [];

        foreach ($expectedParams as $param) {
            if (!array_key_exists($param, $data)) {
                $paramsNotFound[] = $param;
            }
        }

        if ($paramsNotFound) {
            $singular = count($paramsNotFound) === 1;
            $msg = $singular ? "Parameter: " : "Parameters: ";
            foreach ($paramsNotFound as $param) {
                $msg .= $param . ', ';
            }
            $msg = rtrim($msg, ', ');
            $msg .= $singular ? ' is required.' : ' are required.';

            $this->logger->error($msg);
            throw new ApiException($msg);
        }
    }

    /**
     * @param string $endpoint
     */
    private function getUrlEndpoint($endpoint)
    {
        $apiUrl = self::TEST_API_URL;

        if ($this->getConfigValue('payment/instantcredit/environment') == true) {
            $apiUrl = self::PRODUCTION_API_URL;
        }

        return $apiUrl . $endpoint;
    }

    /**
     * @return mixed
     */
    public function getSecretKey() {
        return $this->getConfigValue('payment/instantcredit/secret_key');
    }

    /**
     * @param array $requestData
     * @return array
     */
    public function prepareRequestData($requestData) {

        try {

            $secretKey = $this->getSecretKey();
            if(!$secretKey) {
                throw new Exception('Secret Key is not defined');
            }

            $data = $this->serializer->serialize($requestData);

        } catch (\Exception $e) {

            $this->logger->error('Data:' . $requestData);
            $this->logger->error($e->getMessage());

            return [];
        }

        $signature = hash('sha256', $secretKey . $data);
		$this->logger->error('Signature:' . $signature);
        $requestData = [
            'data' => $data,
            'signature' => $signature
        ];

        return $requestData;

    }

    /**
     * @param $requestUrl
     * @param $data
     * @return array[]
     * @throws \Instant\Credit\Api\ApiException
     */
    public function request($requestUrl, $data)
    {

        $requestData = $this->prepareRequestData($data);

        $this->logger->info(
            'STARTING REQUEST. Endpoint: ' . $requestUrl . ' ~ Body: ' . $this->serializer->serialize($data) .
            ' ~ Signature: ' . $requestData['signature']);

        try {

            /** @var bool $checkSignature Need to check signature from response */
            $checkSignature = false;

            if ($requestUrl !== \Instant\Credit\Api\Data\TokenInterface::API_ENDPOINT) {

                $token = $this->getToken();
                if(!$token) {
                    throw new ApiException('Can\'t get token');
                }

                $this->curl->addHeader("Authorization", "Bearer " . $token);
                $this->curl->addHeader("Signature", $requestData['signature']);

                $checkSignature = true;
            }

            $requestUrl = $this->getUrlEndpoint($requestUrl);

            $this->curl->addHeader("Content-Type", "application/json");

            $this->curl->post($requestUrl, $requestData['data']);

            if ($this->curl->getStatus() === 500) {
                throw new ApiException('API Responds with status 500');
            }

            $response = $this->serializer->unserialize($this->curl->getBody());
            if(!$response) {
                throw new ApiException('There was an error while processing your request.');
            }

            if ($checkSignature) {
                if ($this->curl->getHeaders()['Signature'] !== $this->getSignature($this->curl->getBody())) {
                    throw new ApiException('Signature from response is not equal to calculated signature');
                }
            }


            $this->logger->info(
                'FINISH REQUEST. Endpoint: ' . $requestUrl . ' ~ Body response :  ' .
                $this->serializer->serialize($response)
            );

            return $this->createSuccessResponse($response);

        } catch (\Exception $e) {

            $this->logger->error('Request URL: ' . $requestUrl);
            $this->logger->error('Request Data: ' . $this->serializer->serialize($data));
            $this->logger->error($e->getMessage());

            return $this->createErrorResponse($e->getMessage());
        }
    }

    /**
     * @param $data
     * @return string
     */
    public function getSignature($data)
    {
        $secretKey = $this->scopeConfig->getValue('payment/instantcredit/secret_key', \Magento\Store\Model\ScopeInterface::SCOPE_STORE);
        $hash = hash('sha256', $secretKey . $data);
        return $hash;
    }

    /**
     * Get auth token
     * @return mixed
     * @throws \Instant\Credit\Api\ApiException
     */
    public function getToken()
    {

        /** @var \Instant\Credit\Model\Api\Data\TokenApi $getToken */
        $getToken = $this->token->create()
            ->setClientId($this->getConfigValue('payment/instantcredit/client_id'))
            ->setUsername($this->getConfigValue('payment/instantcredit/username'))
            ->setPassword($this->getConfigValue('payment/instantcredit/password'))
            ->setGrantType('password');

        $tokenRequest = $this->request(
            \Instant\Credit\Api\Data\TokenInterface::API_ENDPOINT,
            $getToken->getData()
        );

        if ($tokenRequest['response']['code'] !== AbstractApi::API_ERROR_CODE_SUCCESS) {
            throw new \Instant\Credit\Api\ApiException(
                'Error obtaining token from Instant Credit \n' .
                'Endpoint: ' . \Instant\Credit\Api\Data\TokenInterface::API_ENDPOINT . ' \n' .
                'Request: ' . $this->serializer->serialize($getToken->getData())
            );
        }

        return $tokenRequest['response']['data']['accessToken'];

    }

}
