<?php

namespace Instant\Credit\Controller\Callback;

use Instant\Credit\Api\ApiException;
use Magento\Framework\App\Action\Action;
use Magento\Framework\App\ActionInterface;
use Magento\Framework\App\CsrfAwareActionInterface;
use Magento\Framework\App\RequestInterface;
use Magento\Framework\App\Request\InvalidRequestException;
use Instant\Credit\Model\Enum\ResultType;
use Magento\Sales\Model\Order;
use Magento\Framework\App\Action\Context;
use Instant\Credit\Model\Api\RequestApiFactory;
use Magento\Framework\App\Request\Http;
use Instant\Credit\Api\Data\CreditConfirmationInterfaceFactory;
use Magento\Framework\Serialize\Serializer\Json;
use Magento\Framework\Controller\Result\JsonFactory;
use Instant\Credit\Model\ResourceModel\Transactions\CollectionFactory;
use Instant\Credit\Model\ResourceModel\Transactions;
use Instant\Credit\Api\Data\TransactionsInterfaceFactory;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Sales\Model\OrderNotifier;
use Magento\Framework\Controller\ResultFactory;
use Magento\Sales\Api\OrderManagementInterface;
use Magento\Sales\Model\Order\Email\Sender\OrderSender;
use Instant\Credit\Logger\InstantCredit;

/**
 * Class Index
 * @package Instant\Credit\Controller\Success
 * @author Orenyi
 */
class Index extends Action implements ActionInterface, CsrfAwareActionInterface
{

    /**
     * Error codes
     */
    const ERROR_SIGNATURE = 'IC1000';
    const ERROR_PARAMS = 'IC1001';
    const ERROR_HEADERS = 'IC1002';
    const ORDER_NOT_FOUND = 'IC1003';
    const ERROR_SECRET_KEY = 'IC1004';
    const ERROR_ALREADY_CONFIRMED = 'IC1005';

    const CUSTOM_ERROR_MSG = 'There was an error processing your request';

    /**
     * @var \Magento\Framework\App\Action\Context
     */
    protected $context;
    /**
     * @var \Instant\Credit\Model\Api\RequestApiFactory
     */
    protected $requestApi;

    /**
     * @var \Magento\Framework\App\Request\Http
     */
    protected $postRequest;

    /**
     * @var \Instant\Credit\Api\Data\CreditConfirmationInterfaceFactory
     */
    protected $creditConfirmation;

    /**
     * @var \Magento\Framework\Serialize\Serializer\Json
     */
    protected $serializer;
    /**
     * @var \Magento\Framework\Controller\Result\JsonFactory
     */
    private $resultJsonFactory;
    /**
     * @var \Instant\Credit\Model\ResourceModel\Transactions\CollectionFactory
     */
    private $transactionsCollection;
    /**
     * @var \Instant\Credit\Model\ResourceModel\Transactions
     */
    private $transactionResource;
    /**
     * @var \Instant\Credit\Api\Data\TransactionsInterfaceFactory
     */
    private $transactionInterfaceFactory;
    /**
     * @var \Magento\Sales\Api\OrderRepositoryInterface
     */
    private $orderRepository;
    /**
     * @var \Magento\Sales\Model\OrderNotifier
     */
    private $orderNotifier;
    /**
     * @var \Magento\Framework\Controller\ResultFactory
     */
    protected $resultFactory;
    /**
     * @var \Magento\Sales\Api\OrderManagementInterface
     */
    private $orderManagement;
    /**
     * @var OrderSender
     */
    protected $orderSender;
    /**
     * @var \Instant\Credit\Logger\InstantCredit
     */
    private $logger;
    /**
     * @var \Magento\Framework\Api\SearchCriteriaBuilder
     */
    private $searchCriteriaBuilder;

    /**
     * CreditConfirmation constructor.
     * @param \Instant\Credit\Model\Api\RequestApiFactory $requestApi
     * @param \Magento\Framework\App\Request\Http $postRequest
     * @param \Instant\Credit\Api\Data\CreditConfirmationInterfaceFactory $creditConfirmation
     * @param \Magento\Framework\Serialize\Serializer\Json $serializer
     * @param \Magento\Framework\Controller\Result\JsonFactory $resultJsonFactory
     * @param \Instant\Credit\Model\ResourceModel\Transactions\CollectionFactory $transactionsCollection
     * @param \Instant\Credit\Model\ResourceModel\Transactions $transactionResource
     * @param \Instant\Credit\Api\Data\TransactionsInterfaceFactory $transactionInterfaceFactory
     * @param \Magento\Sales\Api\OrderRepositoryInterface $orderRepository
     * @param \Magento\Framework\Api\SearchCriteriaBuilder $searchCriteriaBuilder
     * @param \Magento\Sales\Model\OrderNotifier $orderNotifier
     * @param \Magento\Framework\Controller\ResultFactory $resultFactory
     * @param \Magento\Sales\Api\OrderManagementInterface $orderManagement
     * @param \Instant\Credit\Logger\InstantCredit $logger
     */
    public function __construct(
        Context $context,
        RequestApiFactory $requestApi,
        Http $postRequest,
        CreditConfirmationInterfaceFactory $creditConfirmation,
        Json $serializer,
        JsonFactory $resultJsonFactory,
        CollectionFactory $transactionsCollection,
        Transactions $transactionResource,
        TransactionsInterfaceFactory $transactionInterfaceFactory,
        OrderRepositoryInterface $orderRepository,
        SearchCriteriaBuilder $searchCriteriaBuilder,
        OrderNotifier $orderNotifier,
        ResultFactory $resultFactory,
        OrderManagementInterface $orderManagement,
        OrderSender $orderSender,
        InstantCredit $logger
    ) {
        parent::__construct($context);
        $this->requestApi = $requestApi;
        $this->postRequest = $postRequest;
        $this->creditConfirmation = $creditConfirmation;
        $this->serializer = $serializer;
        $this->resultJsonFactory = $resultJsonFactory;
        $this->transactionsCollection = $transactionsCollection;
        $this->transactionResource = $transactionResource;
        $this->transactionInterfaceFactory = $transactionInterfaceFactory;
        $this->orderRepository = $orderRepository;
        $this->orderNotifier = $orderNotifier;
        $this->resultFactory = $resultFactory;
        $this->orderManagement = $orderManagement;
        $this->orderSender = $orderSender;
        $this->logger = $logger;
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;
    }

    /**
     * @return string
     * @throws ApiException
     */
    public function execute()
    {
        $requestApi = $this->requestApi->create();
        $jsonResult = $this->resultJsonFactory->create();

        try {

            if (!$this->postRequest->isPost()) {
                $redirect = $this->resultFactory->create(\Magento\Framework\Controller\ResultFactory::TYPE_REDIRECT);
                $redirect->setUrl('/');
                return $redirect;
            }

            $this->logger->info('START Callback request');

            $contentType = $this->postRequest->getHeader('Content-Type');
            if ($contentType !== 'application/json') {
                throw new \Exception(self::CUSTOM_ERROR_MSG . ' ' . self::ERROR_HEADERS);
            }

            $requestSignature = $this->postRequest->getHeader('Signature');
            if (!$requestSignature) {
                throw new \Exception(self::CUSTOM_ERROR_MSG . ' ' . self::ERROR_HEADERS);
            }

            $postParams = $this->postRequest->getContent();
            if(!$postParams) {
                throw new \Exception(self::CUSTOM_ERROR_MSG . ' ' . self::ERROR_PARAMS);
            }

            $data = $this->serializer->unserialize($postParams);

            $check = $this->checkRequiredParams($data);
            if (!$check) {
                throw new \Exception(self::CUSTOM_ERROR_MSG . ' ' . self::ERROR_PARAMS);
            }

            if (!$requestApi->getSecretKey()) {
                throw new \Exception(self::CUSTOM_ERROR_MSG . ' ' . self::ERROR_SECRET_KEY);
            }

            $signatureCalc = $signature = hash('sha256', $requestApi->getSecretKey() . $postParams);

            // We calculate signature with body and our secret key to check if is correct
            if ($signatureCalc !== $requestSignature) {
                   throw new \Exception('There was an error while processing your request. ' . self::ERROR_SIGNATURE);
            }

            /** @var \Instant\Credit\Model\Api\Data\CreditConfirmationApi $ccResponse */
            $ccResponse = $this->creditConfirmation->create()
                ->setOrderId($data['orderId'])
                ->setTransactionId($data['transactionId'])
                ->setResult('OK')
                ->setResultType($data['resultType'])
                ->setResultDescription('Acknowledgement');

            $getSignature = $requestApi->prepareRequestData($ccResponse->getData())['signature'];
            if (!isset($getSignature)) {
                throw new \Exception('There was an error while processing your request. ' . self::ERROR_SIGNATURE);
            }

            $jsonResult->setData($ccResponse->getData());
            $jsonResult->setHeader('Signature', $getSignature);

            /** @var \Instant\Credit\Model\Api\Data\Transactions $transaction */
            $transaction = $this->getTransactionByOrderId($ccResponse->getOrderId());

            // Not confirmed yet
            if (!$transaction->getEntityId()) {

                $orderId = $data['orderId'];
                $orderSplit = explode('_', $orderId);

                if (count($orderSplit) === 2) { // Order ID must be orderIncrementID_storeId
                    $orderId = $orderSplit[0];
                } else {
                    throw new \Exception('There was an error while processing your request. ' . self::ORDER_NOT_FOUND);
                }

                $searchCriteria = $this->searchCriteriaBuilder
                    ->addFilter('increment_id', $orderId)
                    ->create();
                /** @var \Magento\Sales\Model\Order $order */
                $order = $this->orderRepository->getList($searchCriteria)->getItems();

                if (count($order) === 0) {
                    throw new \Exception('Can\'t find order with Increment ID : ' . $orderId);
                }

                $order = reset($order);

                $order->addCommentToStatusHistory($this->getCommentByResultType($ccResponse->getResultType()));
                $this->orderRepository->save($order);

                if (strtoupper($ccResponse->getResultType()) === ResultType::SUCCESSFUL) {
                    $order->setState('processing')->setStatus('processing');
                    //$this->orderNotifier->notify($order); // Send confirmation order mail
                    //$this->orderSender->send($order, true); // Send confirmation order mail
                    $this->orderRepository->save($order);
                    // send Order email
                    try {
                        $this->orderSender->send($order, true);
                    } catch (\Exception $e) {
                        $this->logger->critical($e);
                    }
                } else {
                    $this->orderManagement->cancel($order->getId());
                }
				$this->logger->info('TransactionId Callback: '.intval($ccResponse->getTransactionId()));
				$this->logger->info('ResultDescription Callback: '.$ccResponse->getResultDescription());
				$this->logger->info('RequestSignature Callback: '.$requestSignature);
				$this->logger->info('Currency Callback: '.$data['currency']);
				$this->logger->info('Amount Callback: '.$data['amount']);
				$this->logger->info('Result Callback: '.$ccResponse->getResult());
				$this->logger->info('ResultType Callback: '.$ccResponse->getResultType());
				$this->logger->info('OrderId Callback: '.$orderId);
				$this->logger->info('Country Callback: '.$data['country']);
				$this->logger->info('START Callback transaction interface');
                /** @var \Instant\Credit\Model\Api\Data\Transactions $transaction */
                $transaction = $this->transactionInterfaceFactory->create();
				$transaction->setTransactionId(intval($ccResponse->getTransactionId()));
				$transaction->setDescription($ccResponse->getResultDescription());
				$transaction->setSignature($requestSignature);
				$transaction->setCurrency($data['currency']);
				$transaction->setAmount($data['amount']);
				$transaction->setResult($ccResponse->getResult());
				$transaction->setResultType($ccResponse->getResultType());
				$transaction->setOrderId($orderId);
				$transaction->setCountry($data['country']);
				$transaction->setConfirmed(true);
				$this->logger->info('Before save Callback transaction interface');
                $this->transactionResource->save($transaction);
				
				$this->logger->info('Save Callback transaction interface');
            }

        } catch (\Exception $e) {
            $this->logger->error('Can\'t create transaction in Callback. ERROR: ' . $e->getMessage());

            $ccResponse = $this->creditConfirmation->create()
                ->setOrderId(isset($data['orderId']) ? $data['orderId'] : '')
                ->setTransactionId(isset($data['transactionId']) ? $data['transactionId'] : '')
                ->setResult('FAIL')
                ->setResultType(isset($data['resultType']) ? $data['resultType'] : '')
                ->setResultDescription($e->getMessage());
            $jsonResult->setData($ccResponse->getData());
        }

        $this->logger->info('END Callback request');

        return $jsonResult;
    }

    /**
     * Check if all params required are sent
     * @param $data
     * @return bool
     */
    public function checkRequiredParams($data)
    {
        $requiredParams = [
            'transactionId',
            'orderId',
            'amount',
            'currency',
            'country',
            'result',
            'resultType',
            'resultDescription'
        ];

        foreach ($requiredParams as $required)
        {
            if (!isset($data[$required])) {
                return false;
            }
        }

        return true;
    }

    /**
     * @param $resultType
     * @return string
     */
    private function getStatusByResultType($resultType)
    {
        if (strtoupper($resultType) === ResultType::SUCCESSFUL) {
            return 'processing';
        }

        return 'canceled';
    }

    /**
     * @param $resultType
     * @return \Magento\Framework\Phrase|string
     */
    private function getCommentByResultType($resultType)
    {
        $deniedText = __('Instant Credit Notification: Financing Canceled');

        switch (strtoupper($resultType)) {
            case ResultType::SUCCESSFUL:
                $comment = __('Instant Credit Notification: Financing Approved');
                break;
            case ResultType::TIMEOUT:
                $comment = $deniedText . ' - ' . 'Timeout';
                break;
            case ResultType::CANCELLED_BY_ENDUSER:
                $comment = $deniedText . ' - ' . __('Cancelled by user');
                break;
            case ResultType::INCORRECT_DATE:
                $comment = $deniedText . ' - ' . __('Incorrect date');
                break;
            case ResultType::ORDER_ID_ALREADY_EXISTS:
                $comment = $deniedText . ' - ' . __('Order ID already exists');
                break;
            case ResultType::NO_LENDERS_AVAILABLE:
                $comment = $deniedText . ' - ' . __('No Lenders available');
                break;
            case ResultType::INCORRECT_CURRENCY:
                $comment = $deniedText . ' - ' . __('Incorrect currency');
                break;
            case ResultType::INCORRECT_COUNTRY:
                $comment = $deniedText . ' - ' . __('Incorrect country');
                break;
            case ResultType::UNDEFINED:
                $comment = $deniedText . ' - ' . __('Undefined');
                break;
            default:
                $comment = $deniedText;
                break;
        }

        return $comment;
    }


    /**
     * @param $orderId
     * @return \Magento\Framework\DataObject
     */
    private function getTransactionByOrderId($orderId)
    {
        return $this->transactionsCollection->create()
            ->addFieldToFilter('order_id', $orderId)
            ->getFirstItem();
    }


    /**
     * @param RequestInterface $request
     * @return InvalidRequestException|null
     */
    public function createCsrfValidationException(RequestInterface $request): ? InvalidRequestException
    {
        return null;
    }

    /**
     * @param RequestInterface $request
     * @return bool|null
     */
    public function validateForCsrf(RequestInterface $request): ?bool
    {
        return true;
    }
}